/*! \file */
/* ************************************************************************
* Copyright (C) 2022 Advanced Micro Devices, Inc. All rights Reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
* ************************************************************************ */
#pragma once

#include <iostream>
#include <sstream>

#include "program_options.hpp"
#include "utility.hpp"

#include "rocalution_bench_solver_parameters.hpp"

//
// @brief Class as an extension of Arguments to deal with the rocalution_bench.
//
struct rocalution_arguments_config : rocalution_bench_solver_parameters
{

    char        precision;
    char        indextype;
    std::string function;
    int         device_id;

    //
    // @brief Default constructor.
    //
    rocalution_arguments_config();

    //
    // @brief Set the description.
    //
    void set_description(options_description& desc);

    //
    // @brief Parse command line arguments.
    //
    int parse(int& argc, char**& argv, options_description& desc);

    //
    // @brief Parse command line arguments with no default value.
    //
    int parse_no_default(int& argc, char**& argv, options_description& desc);
};
